@@ -30,13 +30,24 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 #include <windows.h>
 #include <winsock2.h>
 #define sleep(x) Sleep(x*1000)
+#define EPROTO 71
+#define EBADMSG 77
 #else
 #include <sys/socket.h>
 #include <arpa/inet.h>
+#include <pthread.h>
 #endif
+
+#ifdef HAVE_INOTIFY
+#include <sys/inotify.h>
+#define EVENT_SIZE  (sizeof (struct inotify_event))
+#define EVENT_BUF_LEN (1024 * (EVENT_SIZE + 16))
+#define USBMUXD_DIRNAME "/var/run"
+#define USBMUXD_SOCKET_NAME "usbmuxd"
+#endif /* HAVE_INOTIFY */
+
 #include <unistd.h>
 #include <signal.h>
-#include <pthread.h>
 
 #ifdef HAVE_PLIST
 #include <plist/plist.h>
@@ -56,7 +67,12 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 
 static struct collection devices;
 static usbmuxd_event_cb_t event_cb = NULL;
+#ifdef WIN32
+HANDLE devmon = NULL;
+#else
 pthread_t devmon;
+pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
+#endif
 static int listenfd = -1;
 
 static int use_tag = 0;
@@ -174,7 +190,7 @@ static int receive_packet(int sfd, struct usbmuxd_header *header, void **payload
 				char *strval = NULL;
 				plist_get_string_val(n, &strval);
 				if (strval) {
-					strcpy(dev->serial_number, strval);
+					strncpy(dev->serial_number, strval, 255);
 					free(strval);
 				}
 				n = plist_dict_get_item(props, "LocationID");
@@ -374,19 +390,10 @@ static void generate_event(usbmuxd_event_cb_t callback, const usbmuxd_device_inf
 	callback(&ev, user_data);
 }
 
-/**
- * Tries to connect to usbmuxd and wait if it is not running.
- * 
- * TODO inotify support should come here
- */
-static int usbmuxd_listen()
+static int usbmuxd_listen_poll()
 {
 	int sfd;
-	uint32_t res = -1;
 
-#ifdef HAVE_PLIST
-retry:
-#endif
 	sfd = connect_usbmuxd_socket();
 	if (sfd < 0) {
 		while (event_cb) {
@@ -397,6 +404,86 @@ retry:
 		}
 	}
 
+	return sfd;
+}
+
+#ifdef HAVE_INOTIFY
+static int usbmuxd_listen_inotify()
+{
+	int inot_fd;
+	int watch_d;
+	int sfd;
+
+	sfd = connect_usbmuxd_socket();
+	if (sfd >= 0)
+		return sfd;
+
+	sfd = -1;
+	inot_fd = inotify_init ();
+	if (inot_fd < 0) {
+		fprintf (stderr, "Failed to setup inotify\n");
+		return -2;
+	}
+
+	/* inotify is setup, listen for events that concern us */
+	watch_d = inotify_add_watch (inot_fd, USBMUXD_DIRNAME, IN_CREATE);
+	if (watch_d < 0) {
+		fprintf (stderr, "Failed to setup watch descriptor for socket dir\n");
+		close (inot_fd);
+		return -2;
+	}
+
+	while (1) {
+		ssize_t len, i;
+		char buff[EVENT_BUF_LEN] = {0};
+
+		i = 0;
+		len = read (inot_fd, buff, EVENT_BUF_LEN -1);
+		if (len < 0)
+			goto end;
+		while (i < len) {
+			struct inotify_event *pevent = (struct inotify_event *) & buff[i];
+
+			/* check that it's ours */
+			if (pevent->mask & IN_CREATE &&
+			    pevent->len &&
+			    pevent->name != NULL &&
+			    strcmp(pevent->name, USBMUXD_SOCKET_NAME) == 0) {
+				sfd = connect_usbmuxd_socket ();
+				goto end;
+			}
+			i += EVENT_SIZE + pevent->len;
+		}
+	}
+
+end:
+	inotify_rm_watch(inot_fd, watch_d);
+	close(inot_fd);
+
+	return sfd;
+}
+#endif /* HAVE_INOTIFY */
+
+/**
+ * Tries to connect to usbmuxd and wait if it is not running.
+ */
+static int usbmuxd_listen()
+{
+	int sfd;
+	uint32_t res = -1;
+
+#ifdef HAVE_PLIST
+retry:
+#endif
+
+#ifdef HAVE_INOTIFY
+	sfd = usbmuxd_listen_inotify();
+	if (sfd == -2)
+		sfd = usbmuxd_listen_poll();
+#else
+	sfd = usbmuxd_listen_poll();
+#endif
+
 	if (sfd < 0) {
 		fprintf(stderr, "%s: ERROR: usbmuxd was supposed to be running here...\n", __func__);
 		return sfd;
@@ -440,6 +527,7 @@ int get_next_event(int sfd, usbmuxd_event_cb_t callback, void *user_data)
 		FOREACH(usbmuxd_device_info_t *dev, &devices) {
 			generate_event(callback, dev, UE_DEVICE_REMOVE, user_data);
 			collection_remove(&devices, dev);
+			free(dev);
 		} ENDFOREACH
 		return -EIO;
 	}
@@ -460,8 +548,12 @@ int get_next_event(int sfd, usbmuxd_event_cb_t callback, void *user_data)
 
 		devinfo->handle = dev->device_id;
 		devinfo->product_id = dev->product_id;
-		memset(devinfo->uuid, '\0', sizeof(devinfo->uuid));
-		memcpy(devinfo->uuid, dev->serial_number, sizeof(devinfo->uuid));
+		memset(devinfo->udid, '\0', sizeof(devinfo->udid));
+		memcpy(devinfo->udid, dev->serial_number, sizeof(devinfo->udid));
+
+		if (strcasecmp(devinfo->udid, "ffffffffffffffffffffffffffffffffffffffff") == 0) {
+			sprintf(devinfo->udid + 32, "%08x", devinfo->handle);
+		}
 
 		collection_add(&devices, devinfo);
 		generate_event(callback, devinfo, UE_DEVICE_ADD, user_data);
@@ -477,8 +569,9 @@ int get_next_event(int sfd, usbmuxd_event_cb_t callback, void *user_data)
 		} else {
 			generate_event(callback, devinfo, UE_DEVICE_REMOVE, user_data);
 			collection_remove(&devices, devinfo);
+			free(devinfo);
 		}
-	} else {
+	} else if (hdr.length > 0) {
 		fprintf(stderr, "%s: Unexpected message type %d length %d received!\n", __func__, hdr.message, hdr.length);
 	}
 	if (payload) {
@@ -487,6 +580,14 @@ int get_next_event(int sfd, usbmuxd_event_cb_t callback, void *user_data)
 	return 0;
 }
 
+static void device_monitor_cleanup(void* data)
+{
+	collection_free(&devices);
+
+	close_socket(listenfd);
+	listenfd = -1;
+}
+
 /**
  * Device Monitor thread function.
  *
@@ -496,6 +597,9 @@ static void *device_monitor(void *data)
 {
 	collection_init(&devices);
 
+#ifndef WIN32
+	pthread_cleanup_push(device_monitor_cleanup, NULL);
+#endif
 	while (event_cb) {
 
 		listenfd = usbmuxd_listen();
@@ -511,8 +615,11 @@ static void *device_monitor(void *data)
 		}
 	}
 
-	collection_free(&devices);
-
+#ifndef WIN32
+	pthread_cleanup_pop(1);
+#else
+	device_monitor_cleanup();
+#endif
 	return NULL;
 }
 
@@ -525,7 +632,15 @@ int usbmuxd_subscribe(usbmuxd_event_cb_t callback, void *user_data)
 	}
 	event_cb = callback;
 
+#ifdef WIN32
+	res = 0;
+	devmon = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)device_monitor, user_data, 0, NULL);
+	if (devmon == NULL) {
+		res = GetLastError();
+	}
+#else
 	res = pthread_create(&devmon, NULL, device_monitor, user_data);
+#endif
 	if (res != 0) {
 		fprintf(stderr, "%s: ERROR: Could not start device watcher thread!\n", __func__);
 		return res;
@@ -537,12 +652,18 @@ int usbmuxd_unsubscribe()
 {
 	event_cb = NULL;
 
+	shutdown_socket(listenfd, SHUT_RDWR);
+
+#ifdef WIN32
+	if (devmon != NULL) {
+		WaitForSingleObject(devmon, INFINITE);
+	}
+#else
 	if (pthread_kill(devmon, 0) == 0) {
-		close_socket(listenfd);
-		listenfd = -1;
-		pthread_kill(devmon, SIGINT);
+		pthread_cancel(devmon);
 		pthread_join(devmon, NULL);
 	}
+#endif
 
 	return 0;
 }
@@ -552,12 +673,15 @@ int usbmuxd_get_device_list(usbmuxd_device_info_t **device_list)
 	int sfd;
 	int listen_success = 0;
 	uint32_t res;
+	struct collection tmpdevs;
 	usbmuxd_device_info_t *newlist = NULL;
 	struct usbmuxd_header hdr;
-	struct usbmuxd_device_record *dev_info;
+	struct usbmuxd_device_record *dev;
 	int dev_cnt = 0;
 	void *payload = NULL;
 
+	*device_list = NULL;
+
 #ifdef HAVE_PLIST
 retry:
 #endif
@@ -593,32 +717,46 @@ retry:
 		return -1;
 	}
 
-	*device_list = NULL;
+	collection_init(&tmpdevs);
+
 	// receive device list
 	while (1) {
 		if (receive_packet(sfd, &hdr, &payload, 1000) > 0) {
 			if (hdr.message == MESSAGE_DEVICE_ADD) {
-				dev_info = payload;
-				newlist = (usbmuxd_device_info_t *) realloc(*device_list, sizeof(usbmuxd_device_info_t) * (dev_cnt + 1));
-				if (newlist) {
-					newlist[dev_cnt].handle =
-						(int) dev_info->device_id;
-					newlist[dev_cnt].product_id =
-						dev_info->product_id;
-					memset(newlist[dev_cnt].uuid, '\0',
-						   sizeof(newlist[dev_cnt].uuid));
-					memcpy(newlist[dev_cnt].uuid,
-						   dev_info->serial_number,
-						   sizeof(newlist[dev_cnt].uuid));
-					*device_list = newlist;
-					dev_cnt++;
-				} else {
-					fprintf(stderr,
-						"%s: ERROR: out of memory when trying to realloc!\n",
-						__func__);
-					if (payload)
-						free(payload);
-					break;
+				dev = payload;
+				usbmuxd_device_info_t *devinfo = (usbmuxd_device_info_t*)malloc(sizeof(usbmuxd_device_info_t));
+				if (!devinfo) {
+					fprintf(stderr, "%s: Out of memory!\n", __func__);
+					free(payload);
+					return -1;
+				}
+
+				devinfo->handle = dev->device_id;
+				devinfo->product_id = dev->product_id;
+				memset(devinfo->udid, '\0', sizeof(devinfo->udid));
+				memcpy(devinfo->udid, dev->serial_number, sizeof(devinfo->udid));
+
+				if (strcasecmp(devinfo->udid, "ffffffffffffffffffffffffffffffffffffffff") == 0) {
+					sprintf(devinfo->udid + 32, "%08x", devinfo->handle);
+				}
+
+				collection_add(&tmpdevs, devinfo);
+
+			} else if (hdr.message == MESSAGE_DEVICE_REMOVE) {
+				uint32_t handle;
+				usbmuxd_device_info_t *devinfo = NULL;
+
+				memcpy(&handle, payload, sizeof(uint32_t));
+
+				FOREACH(usbmuxd_device_info_t *di, &tmpdevs) {
+					if (di && di->handle == handle) {
+						devinfo = di;
+						break;
+					}
+				} ENDFOREACH
+				if (devinfo) {
+					collection_remove(&tmpdevs, devinfo);
+					free(devinfo);
 				}
 			} else {
 				fprintf(stderr, "%s: Unexpected message %d\n", __func__, hdr.message);
@@ -635,9 +773,19 @@ retry:
 	// explicitly close connection
 	close_socket(sfd);
 
-	// terminating zero record
-	newlist = (usbmuxd_device_info_t*) realloc(*device_list, sizeof(usbmuxd_device_info_t) * (dev_cnt + 1));
-	memset(newlist + dev_cnt, 0, sizeof(usbmuxd_device_info_t));
+	// create copy of device info entries from collection
+	newlist = (usbmuxd_device_info_t*)malloc(sizeof(usbmuxd_device_info_t) * (collection_count(&tmpdevs) + 1));
+	dev_cnt = 0;
+	FOREACH(usbmuxd_device_info_t *di, &tmpdevs) {
+		if (di) {
+			memcpy(&newlist[dev_cnt], di, sizeof(usbmuxd_device_info_t));
+			free(di);
+			dev_cnt++;
+		}
+	} ENDFOREACH
+	collection_free(&tmpdevs);
+
+	memset(&newlist[dev_cnt], 0, sizeof(usbmuxd_device_info_t));
 	*device_list = newlist;
 
 	return dev_cnt;
@@ -651,7 +799,7 @@ int usbmuxd_device_list_free(usbmuxd_device_info_t **device_list)
 	return 0;
 }
 
-int usbmuxd_get_device_by_uuid(const char *uuid, usbmuxd_device_info_t *device)
+int usbmuxd_get_device_by_udid(const char *udid, usbmuxd_device_info_t *device)
 {
 	usbmuxd_device_info_t *dev_list = NULL;
 
@@ -665,17 +813,17 @@ int usbmuxd_get_device_by_uuid(const char *uuid, usbmuxd_device_info_t *device)
 	int i;
 	int result = 0;
 	for (i = 0; dev_list[i].handle > 0; i++) {
-	 	if (!uuid) {
+	 	if (!udid) {
 			device->handle = dev_list[i].handle;
 			device->product_id = dev_list[i].product_id;
-			strcpy(device->uuid, dev_list[i].uuid);
+			strcpy(device->udid, dev_list[i].udid);
 			result = 1;
 			break;
 		}
-		if (!strcmp(uuid, dev_list[i].uuid)) {
+		if (!strcmp(udid, dev_list[i].udid)) {
 			device->handle = dev_list[i].handle;
 			device->product_id = dev_list[i].product_id;
-			strcpy(device->uuid, dev_list[i].uuid);
+			strcpy(device->udid, dev_list[i].udid);
 			result = 1;
 			break;
 		}
